/*
 * log_dump
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <glib.h>
#include <gio/gio.h>
#include "log_dump.h"
#include "dbus-handler.h"

/* Dbus activation */
#define CRASH_BUS_NAME          "org.tizen.system.crash"
#define CRASH_OBJECT_PATH       "/Org/Tizen/System/Crash/Crash"

/* Log dump signal */
#define LOG_DUMP_BUS_NAME       "org.tizen.system.logdump"
#define LOG_DUMP_OBJECT_PATH    "/Org/Tizen/System/LogDump"
#define LOG_DUMP_INTERFACE_NAME LOG_DUMP_BUS_NAME
#define LOG_DUMP_START_SIGNAL   "Start"
#define LOG_DUMP_FINISH_SIGNAL  "Finish"

#define TIMEOUT_INTERVAL 30

static GMainLoop *loop;
static GMutex timeout_mutex;
static guint timeout_id;
static GDBusNodeInfo *introspection_data;
static const gchar introspection_xml[] =
"<node>"
" <interface name='org.tizen.system.crash.Crash'>"
"  <method name='dump_log'>"
"   <arg type='s' name='arg' direction='in'/>"
"   <arg type='i' name='response' direction='out'/>"
"  </method>"
"  <method name='delete_dump'>"
"   <arg type='i' name='response' direction='out'/>"
"  </method>"
" </interface>"
"</node>";

static int timeout_cb(gpointer data)
{
	LOGI("Time out!");
	g_main_loop_quit((GMainLoop *)data);

	return 0;
}

static void add_timeout(void)
{
	g_mutex_lock(&timeout_mutex);

	if (timeout_id)
		g_source_remove(timeout_id);
	timeout_id = g_timeout_add_seconds(TIMEOUT_INTERVAL, timeout_cb, loop);

	g_mutex_unlock(&timeout_mutex);
	LOGI("Add loop timeout (%d)", TIMEOUT_INTERVAL);
}

static void remove_timeout(void)
{
	g_mutex_lock(&timeout_mutex);

	if (timeout_id) {
		g_source_remove(timeout_id);
		timeout_id = 0;
	}

	g_mutex_unlock(&timeout_mutex);
	LOGI("Remove loop timeout");
}

static void method_call_handler(GDBusConnection *conn,
				const gchar *sender,
				const gchar *object_path,
				const gchar *iface_name,
				const gchar *method_name,
				GVariant *parameters,
				GDBusMethodInvocation *invocation,
				gpointer user_data)
{
	int ret = -1;
	const gchar *arg;

	remove_timeout();

	if (g_strcmp0(method_name, "dump_log") == 0) {
		g_variant_get(parameters, "(&s)", &arg);
		if (g_strcmp0(arg, "normal") == 0)
			ret = log_dump(OPT_NORMAL);
		else if (g_strcmp0(arg, "short") == 0)
			ret = log_dump(OPT_SHORT);
		else
			LOGE("Wrong option for log_dump");
	} else if (g_strcmp0(method_name, "delete_dump") == 0) {
		ret = delete_dump();
	}

	g_dbus_method_invocation_return_value(invocation,
			g_variant_new("(i)", ret));

	add_timeout();
}

static const GDBusInterfaceVTable interface_vtable = {
	method_call_handler,
	NULL,
	NULL
};

static void on_bus_acquired(GDBusConnection *conn,
			    const gchar *name,
			    gpointer user_data)
{
	guint registration_id;

	registration_id = g_dbus_connection_register_object(conn,
			CRASH_OBJECT_PATH, introspection_data->interfaces[0],
			&interface_vtable, NULL, NULL, NULL);
	if (registration_id == 0)
		LOGE("Failed to g_dbus_connection_register_object");
}

static void on_name_acquired(GDBusConnection *conn,
			     const gchar *name,
			     gpointer user_data)
{
	LOGD("Acquired the name %s on the system bus", name);
}

static void on_name_lost(GDBusConnection *conn,
			 const gchar *name,
			 gpointer user_data)
{
	LOGD("Lost the name %s on the system bus", name);
}

static void dbus_init(void)
{
	GDBusConnection *conn = NULL;
	GError *error = NULL;

	introspection_data =
		g_dbus_node_info_new_for_xml(introspection_xml,	NULL);
	if (introspection_data == NULL) {
		LOGE("Failed to init g_dbus_info_new_for_xml");
		return;
	}

	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (!conn) {
		LOGE("Failed to get dbus");
		return;
	}

	if (error) {
		LOGE("Failed to get dbus: %s", error->message);
		g_error_free(error);
		return;
	}

	g_bus_own_name(G_BUS_TYPE_SYSTEM, CRASH_BUS_NAME,
			G_BUS_NAME_OWNER_FLAGS_NONE, on_bus_acquired,
			on_name_acquired, on_name_lost, NULL, NULL);
}

int log_dump_dbus(void)
{
	loop = g_main_loop_new(NULL, false);

	dbus_init();

	g_mutex_init(&timeout_mutex);
	add_timeout();

	LOGI("log_dump_dbus activated");
	g_main_loop_run(loop);

	if (introspection_data)
		g_dbus_node_info_unref(introspection_data);
	g_mutex_clear(&timeout_mutex);
	g_main_loop_unref(loop);

	return 0;
}

static int broadcast_logdump(const char *signal)
{
	GDBusConnection *conn;
	GError *error = NULL;

	LOGI("broadcast signal: %s", signal);
	conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (error) {
		LOGE("Failed to get dbus: %s", error->message);
		g_error_free(error);
		return -1;
	}

	g_dbus_connection_emit_signal(conn,
				      NULL,
				      LOG_DUMP_OBJECT_PATH,
				      LOG_DUMP_INTERFACE_NAME,
				      signal,
				      NULL,
				      &error);
	if (error) {
		LOGE("Failed to emit signal: %s", error->message);
		g_error_free(error);
		return -1;
	}

	return 0;
}

int broadcast_logdump_start(void)
{
	return broadcast_logdump(LOG_DUMP_START_SIGNAL);
}

int broadcast_logdump_finish(void)
{
	return broadcast_logdump(LOG_DUMP_FINISH_SIGNAL);
}
